#!/usr/bin/env python

from pandas import Series
from math import exp, log
from numpy import nan
from sys import stderr


class Expected1Found0(ValueError): pass


class Expected1FoundN(ValueError):
    def __init__(self, name, num, elems, show_num=10):
        self.name = name
        self.num = num
        self.elems = elems
        self.msg = f'{len(elems)} {name} found: {",".join([str(elem) for elem in list(elems)[:show_num]])}'
        super().__init__(self.msg)


class WrongKey(ValueError):
    def __init__(self, expected, actual):
        self.expected = expected
        self.actual = actual
        self.msg = f'Expected {expected}, actual {actual}'
        super().__init__(self.msg)


class WrongKeys(ValueError):
    def __init__(self, expected, actual):
        self.missing = set(expected).difference(actual)
        self.extra = set(actual).difference(expected)
        self.msg = f'Expected {expected}, actual {actual}. Extra: {self.extra}, missing: {self.missing}'
        super().__init__()


def singleton(elems, fn=None, empty_ok=False, name='elems', dedupe=True, key=None):
    if isinstance(elems, Series):
        elems = elems.unique().tolist()
    elif isinstance(elems, dict):
        if key:
            keys = list(elems.keys())
            actual_key = singleton(keys)
            if actual_key != key:
                raise WrongKey(key, actual_key)
            return elems[key]
        else:
            elems = elems.items()
    if fn:
        elems = [elem for elem in elems if fn(elem)]
    else:
        if dedupe:
            elems = set(elems)
    if not elems:
        if empty_ok:
            return None
        raise Expected1Found0(f'No {name} found')
    if len(elems) > 1:
        raise Expected1FoundN(name, len(elems), elems)
    [elem] = elems
    return elem


def only(obj, *keys):
    if not keys:
        return obj
    [ keys, *rest ] = keys
    key, scalar = None, False
    if isinstance(keys, str):
        key, scalar = keys, True
        keys = [key]
    actual_keys = list(sorted(obj.keys()))
    expected_keys = list(sorted(keys))
    if expected_keys != actual_keys:
        raise WrongKeys(expected_keys, actual_keys)
    if scalar:
        return only(obj[key], *rest)
    else:
        return [ only(obj[k], *rest) for k in keys ]


def coerce(value, choices, ε=1e-2, multi_ok=False, errors='raise', warn=True):
    assert errors in ['raise', 'coerce', 'ignore', ]
    elems = sorted(
        [
            dict(
                ratio=abs(log(value / choice)),
                choice=choice
            )
            for choice in choices
        ],
        key=lambda r: r['ratio']
    )
    log_ε = log(1 + ε)
    valids = [elem for elem in elems if elem['ratio'] <= log_ε]
    best = elems[0]
    choice = best['choice']
    if not valids:
        msg = f'Best choice {choice} for value {value} has error {exp(best["ratio"]) - 1} > {ε}'
        if errors == 'raise':
            raise ValueError(msg)
        if warn:
            stderr.write(msg + '\n')
        if errors == 'coerce':
            return nan
        else:
            return value
    if len(valids) > 1 and not multi_ok:
        raise ValueError(
            '%d choices passed ε<%f filter:\n\t%s' % (
                len(valids),
                ε,
                "\n\t".join([str(r["choice"]) for r in valids]),
            )
        )
    return best['choice']


def is_subsequence(seq, s):
    if not seq:
        return True
    [ ch, *rest ] = seq
    idx = s.find(ch)
    return is_subsequence(rest, s[(idx+1):]) if idx >= 0 else False



from contextlib import contextmanager
from ..process import run
from . import branch as _branch

class Checkout:
    def __call__(checkout, branch, ref=None, create_ok=False, ctx=True, sha_ok=True):
        prev = _branch.current(sha_ok=sha_ok)
        exists = _branch.exists(branch)
        if not exists and not create_ok:
            raise ValueError(f"Branch {branch} doesn't exist")
        if exists and not ref:
            run('git','checkout',branch)
        else:
            run('git','checkout','-B',branch,ref)
        
        if ctx:
            @contextmanager
            def ret():
                try:
                    yield prev
                finally:
                    checkout(prev, ctx=False)
            
            return ret()
        else:
            return prev
    
    def mk(checkout, branch, ref=None, **kwargs):
        return checkout(branch, ref=ref, create_ok=True, **kwargs)




from os.path import exists

import plotly.graph_objects as go

DEFAULT_PLOTS_DIR = 'www/public/plots'
DEFAULT_MARGIN = { 't': 0, 'r': 25, 'b': 0, 'l': 0, }


def save(
        fig, title, name,
        bg=None, hoverx=False, hovertemplate=None, png_title=False,
        yrange='tozero', bottom_legend='all',
        pretty=False, margin=None,
        dir=None, w=None, h=None,
        xtitle=None, ytitle=None, ltitle=None,
        grid=None, xgrid=None, ygrid=None,
        **layout,
):
    """Plotly wrapper with convenience kwargs for common configurations"""
    if not dir:
        if exists(DEFAULT_PLOTS_DIR):
            dir = DEFAULT_PLOTS_DIR
        elif exists(f'../{DEFAULT_PLOTS_DIR}'):
            dir = f'../{DEFAULT_PLOTS_DIR}'
        else:
            dir = '.'

    bottom_legend_kwargs = dict(
        orientation='h',
        x=0.5,
        xanchor='center',
        yanchor='top',
    )
    layout['xaxis_title'] = layout.get('xaxis_title', xtitle or '')
    layout['yaxis_title'] = layout.get('yaxis_title', ytitle or '')
    layout['legend_title'] = layout.get('legend_title', ltitle or '')
    if bg:
        layout['plot_bgcolor'] = 'white'
        layout['paper_bgcolor'] = 'white'
    if hoverx:
        layout['hovermode'] = 'x'
        fig.update_traces(hovertemplate=hovertemplate)
    elif hovertemplate:
        fig.update_traces(hovertemplate=hovertemplate)
    if yrange:
        layout['yaxis_rangemode'] = yrange
    if bottom_legend == 'all':
        if 'legend' not in layout:
            layout['legend'] = {}
        layout['legend'].update(**bottom_legend_kwargs)
    if xgrid:
        fig.update_xaxes(gridcolor=xgrid,)
    elif grid:
        fig.update_xaxes(gridcolor=grid,)
    if ygrid:
        fig.update_yaxes(gridcolor=ygrid,)
    elif grid:
        fig.update_yaxes(gridcolor=grid,)
    title_layout = dict(title=title, title_x=0.5)
    if png_title:
        layout.update(title_layout)
    fig.update_layout(**layout)
    fig.update_yaxes(rangemode='tozero')
    saved = go.Figure(fig)
    if not png_title:
        # only need to do this if it wasn't already done above
        fig.update_layout(**title_layout)

    if bottom_legend is True:
        saved.update_layout(
            legend=bottom_legend_kwargs,
        )
    saved.update_layout(margin=margin or DEFAULT_MARGIN)
    saved.write_json(f'{dir}/{name}.json', pretty=pretty)
    saved.write_image(f'{dir}/{name}.png', width=w, height=h)

    return fig



#!/usr/bin/env python
# coding: utf-8

# In[ ]:


from ..process import *

def ls():
    return lines('git','submodule','foreach','--quiet','echo $name')

def exists(name): return name in ls()

def add(url, path=None):
    if exists(name): return
    cmd = ['git','submodule','add','url']
    if path: cmd.append(path)
    run(cmd)




#!/usr/bin/env python

# # Context Manager utilities

# General/Common imports:

from contextlib import AbstractContextManager, contextmanager, suppress


# ## `nullcontext`
# Define `nullcontext` in a cross-version way:
try:
    # Python ≥3.7
    from contextlib import nullcontext
except ImportError:
    # Python <3.7
    class nullcontext(object):
        def __enter__(self): pass
        def __exit__(self, *args): pass


# Verify it works:
with nullcontext(): pass


# ## `catch`: context manager for catching+verifying `Exception`s
class catch(AbstractContextManager):
    def __init__(self, *excs):
        self.excs = excs

    def __exit__(self, exc_type, exc_value, traceback):
        if not exc_type:
            if len(self.excs) == 1:
                raise AssertionError(f'No {self.excs[0].__name__} was thrown')
            else:
                raise AssertionError(f'None of {",".join([e.__name__ for e in self.excs])} were thrown')
        
        if not [ isinstance(exc_value, exc) for exc in self.excs ]:
            raise exc_value

        return True


# ## `no`: context manager for verifying `NameError`s (undefined variable names)
no = catch(NameError)


def run_then_raise(fns, vals=None):
    """Run multiple functions, catching all Exceptions and raise-chaining them at the end"""
    if not fns: return vals
    if not vals: vals = []
    (fn, *fns) = fns
    try:
        vals.append(fn())
    except Exception as e:
        run_then_raise(fns, vals=vals)
        raise e
    else:
        return run_then_raise(fns, vals=vals)


def bind_exit(ctx): return lambda: ctx.__exit__(None, None, None)


@contextmanager
def ctxs(ctxs):
    vals = []
    try:
        for ctx in ctxs:
            val = ctx.__enter__()
            vals.append(val)
        yield vals
    except Exception as e:
        run_then_raise(
            reversed([
                bind_exit(ctx)
                for ctx in ctxs[:len(vals)]  # truncate to just the ctxs that were successfully entered
            ])
        )
        raise e
    else:
        fns = [ bind_exit(ctx) for ctx in reversed(ctxs) ]
        run_then_raise(fns)


def contexts(ctxs):
    @contextmanager
    def fn(ctxs):
        if not ctxs:
            yield
        else:
            [ ctx, *rest ] = ctxs
            with ctx as v, fn(rest) as vs:
                yield [ v, *vs ]
    return fn(ctxs)



class cached_property:
    """Based on code from David Beazley's "Python Cookbook" / https://stackoverflow.com/q/62160411/544236."""
    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            value = instance.__dict__[self.func.__name__] = self.func(instance)
            return value



from contextlib import AbstractContextManager
import json
from os.path import exists, join
from os import getcwd, remove
import shlex
from tempfile import NamedTemporaryFile
from types import TracebackType
from typing import Optional, Type

from ..context import nullcontext
from ..process import sh

from .image import Image
from .util import escape


class File(AbstractContextManager):
    _file = None

    def __init__(self, path=None, tag=None, rm=None, copy_dir=None, extend=None, **kwargs):
        if kwargs:
            if 'dir' in kwargs:
                dir = kwargs.pop('dir')
            else:
                dir = getcwd()
            if kwargs:
                raise ValueError(f'Unrecognized kwargs: {kwargs}')
        else:
            dir = getcwd()

        if path:
            self.rm = bool(rm)
        else:
            self.rm = rm != False

        self.path = path
        self.fd = None
        self.tag = tag
        self.dir = dir
        self.copy_dir = copy_dir

        if extend:
            with open(extend, 'r') as f:
                self.lines = [line.rstrip('\n') for line in f.readlines()]
        else:
            self.lines = []

    def __enter__(self):
        if File._file:
            raise RuntimeError(f"Can't enter {self}, already entered {File._file}")
        File._file = self
        return self

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType]
    ):
        File._file = None
        if self.tag:
            self.build(self.tag)
        if self.path and exists(self.path) and self.rm:
            remove(self.path)

    def write(self, *lines):
        self.lines += lines

    def build(self, tag=None, dir=None, **build_args):
        tag = tag or self.tag
        if not tag:
            raise ValueError(f'Missing tag for image build')

        dir = dir or self.dir or '.'
        rm = self.rm
        path = self.path
        mode = 'w'

        if path:
            ctx = nullcontext()
            if exists(path):
                mode = 'a'
        else:
            ctx = NamedTemporaryFile(prefix='Dockerfile.', dir=dir, delete=rm)
            rm = False
            path = ctx.name

        with ctx:
            with open(path, mode) as fd:
                fd.writelines('%s\n' % l for l in self.lines)

            sh(
                'docker', 'build',
                '-f', path,
                '-t', tag,
                [['--build-arg', f'{k}={v}'] for k, v in build_args.items()],
                dir
            )

            if rm:
                remove(path)

        return Image(tag, self)

    def COPY(self, *args, **kwargs):
        (*srcs, dst) = args
        if 'dir' in kwargs:
            dir = kwargs.pop('dir')
            if kwargs:
                raise ValueError(f'Unexpected kwargs: {",".join(kwargs)}')
        else:
            dir = self.copy_dir
        if dir:
            srcs = [join(dir, src) for src in srcs]
        self.write(f'COPY {" ".join([*srcs, dst])}')

    def kvs(self, cmd, *args, **kwargs):
        flattened = []
        for arg in args:
            if isinstance(arg, str):
                flattened.append(arg)
            elif isinstance(arg, tuple):
                if len(arg) != 2:
                    raise RuntimeError(f'Invalid tuple arg (required len 2): %s' % str(arg))
                k, v = arg
                k = escape(str(k))
                v = escape(str(v))
                flattened.append(f'"{k}"="{v}"')
            elif isinstance(arg, dict):
                for k, v in arg.items():
                    flattened.append(f'"{escape(k)}"="{escape(v)}"')
            else:
                raise ValueError(f'Unrecognized argument to {cmd}: {arg}')
        for k, v in kwargs.items():
            flattened.append(f'"{escape(k)}"="{escape(v)}"')
        if flattened:
            self.write(f'{cmd} {" ".join(flattened)}')

    def NOTE(self, *lines):
        self.write(*[f'# {line}' for line in lines])

    def ENV(self, *args, **kwargs):
        return self.kvs('ENV', *args, **kwargs)

    def LABEL(self, *args, **kwargs):
        return self.kvs('LABEL', *args, **kwargs)

    def RUN(self, *cmds):
        if cmds:
            self.write('RUN %s' % ' \\\n && '.join(cmds))

    def ARG(self, k, v=None):
        if v is None:
            self.write(f'ARG {k}')
        else:
            self.write(f'ARG {k}={v}')

    def FROM(self, repo, tag=None, registry=None):
        if registry:
            repo = f'{registry}/{repo}'
        if tag:
            repo = f'{repo}:{tag}'
        self.write(f'FROM {repo}')

    def LN(self):
        self.write('')

    def WORKDIR(self, dst='/'):
        self.write(f'WORKDIR {dst}')

    def ENTRYPOINT(self, *args, shell=False):
        if shell:
            self.write(f'ENTRYPOINT {shlex.join(args)}')
        else:
            self.write(f'ENTRYPOINT {json.dumps(args)}')

    def USER(self, user, group=None):
        if group:
            self.write(f'USER {user}:{group}')
        else:
            self.write(f'USER {user}')



from dataclasses import dataclass
from datetime import datetime as dt, date
from functools import wraps
import re
from math import ceil
from typing import Tuple, Union, Generator

import pandas as pd
from click import option
from pandas.core.tools.datetimes import DatetimeScalar


# Types that can be passed to the Month constructor
Monthy = Union['Month', str, int, None]


@dataclass(init=False, order=True, eq=True, unsafe_hash=True)
class YM:
    y: int
    m: int

    RGX = re.compile(r'(?P<year>\d{4})(?:-?(?P<month>\d\d))?')

    def _init_from_str(self, arg):
        m = self.RGX.fullmatch(arg)
        if not m:
            raise ValueError('Invalid month string: %s' % arg)
        year = int(m['year'])
        month = int(m['month']) if m['month'] else 1
        self.__init__(year, month)

    def _verify(self):
        if not isinstance(self.y, int):
            raise ValueError('Year %s must be int, not %s' % (str(self.y), type(self.y)))
        if not isinstance(self.m, int):
            raise ValueError('Month %s must be int, not %s' % (str(self.m), type(self.m)))

    def _init_now(self):
        now = dt.now()
        self.y = now.year
        self.m = now.month

    def __init__(self, *args, **kwargs):
        if kwargs:
            if args:
                raise ValueError(f'Pass args xor kwargs: {args}, {kwargs}')
            keys = list(kwargs.keys())
            if keys == ['y', 'm']:
                self.y = kwargs['y']
                self.m = kwargs['m']
            else:
                raise ValueError(f"Unrecognized kwargs: {kwargs}")
        elif len(args) == 2:
            self.y, self.m = int(args[0]), int(args[1])
            self._verify()
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, str):
                self._init_from_str(arg)
            elif isinstance(arg, int):
                self._init_from_str(str(arg))
            elif hasattr(arg, 'year') and hasattr(arg, 'month'):
                self.y = arg.y
                self.m = arg.m
                self._verify()
            elif arg is None:
                self._init_now()
            elif 'year' in arg and 'month' in arg:
                self.y = int(arg['year'])
                self.m = int(arg['month'])
                self._verify()
            else:
                raise ValueError('Unrecognized argument: %s' % str(arg))
        elif not args:
            self._init_now()
        else:
            raise ValueError('Unrecognized arguments: %s' % str(args))

    @property
    def year(self):
        return self.y

    @property
    def month(self):
        return self.m

    def str(self, sep=''):
        return '%d%s%02d' % (self.y, sep, self.m)

    def __str__(self):
        return self.str()

    def __int__(self):
        return int(str(self))

    def format(self, url, **kwargs):
        return url.format(ym=str(self), y=str(self.y), m=str(self.m), **kwargs)

    @property
    def dt(self) -> DatetimeScalar:
        return pd.to_datetime('%d-%02d' % (self.y, self.m))

    @property
    def date(self) -> date:
        return self.dt.date()

    @property
    def dates(self) -> Tuple['date', 'date']:
        start = self.date
        end = (self + 1).date
        return start, end

    def __add__(self, n: int) -> 'YM':
        if not isinstance(n, int):
            raise ValueError('%s: can only add an integer to a Month, not %s: %s' % (str(self), str(type(n)), str(n)))
        y, m = self.y, self.m + n - 1
        y += m // 12
        m = (m % 12) + 1
        return YM(y, m)

    def __sub__(self, n: int) -> 'YM':
        if not isinstance(n, int):
            raise ValueError('%s: can only add an integer to a Month, not %s: %s' % (str(self), str(type(n)), str(n)))
        y, m = self.y, self.m - n - 1
        if m <= 0:
            years = int(ceil(-m / 12))
            y -= years
            m += 12 * years
            assert 0 <= m < 12
        m += 1
        return YM(y, m)

    def until(self, end: 'YM' = None, step: int = 1) -> Generator['YM', None, None]:
        cur: YM = YM(self)
        while end is None \
                or (step > 0 and cur < end) \
                or (step < 0 and cur > end):
            yield cur
            cur = cur + step


def dates(default_start=None, default_end=None):
    def _dates(fn):
        @option('-d', '--dates')
        @wraps(fn)
        def _fn(*args, dates=None, **kwargs):
            if dates:
                pcs = dates.split('-')
                if len(pcs) == 2:
                    [ start, end ] = pcs
                    start = YM(start) if start else default_start
                    end = YM(end) if end else default_end
                elif len(pcs) == 1:
                    [ym] = pcs
                    ym = YM(ym)
                    start = ym
                    end = ym + 1
                else:
                    raise ValueError(f"Unrecognized -d/--dates: {dates}")
            else:
                start, end = default_start, default_end
            fn(*args, start=start, end=end, **kwargs)

        return _fn

    return _dates



from dataclasses import dataclass

from typing import Optional, Callable, TypeVar

T = TypeVar('T')


class Unset:
    def __bool__(self):
        return False


@dataclass
class DefaultDict(dict[str, T]):
    """`defaultdict` replacement where `in` reflects the presence of a fallback/default value.

    Can be initialized from a list[str], where each element is either a `k=v` pair or a fallback value (with no `=`)."""
    configs: dict[str, T]
    default: T = Unset

    @staticmethod
    def parse_configs(
            configs: list[str],
            name2value: Optional[Callable[[str], T]] = None,
            fallback: Optional[T] = Unset,
    ) -> (T, dict[str, T]):
        default = fallback
        default_set = False
        kwargs: dict[str, T] = {}
        for write_config in configs:
            kv = write_config.split('=', 1)
            if len(kv) == 2:
                k, name = kv
                value = name2value(name) if name2value else name
                kwargs[k] = value
            elif len(kv) == 1:
                [name] = kv
                if default_set:
                    raise ValueError(f'Multiple defaults found: {default}, {name}')
                default = name2value(name) if name2value else name
                default_set = True
            else:
                raise ValueError(f"Unrecognized defaultdict config arg: {kv}")
        return default, kwargs

    @staticmethod
    def load(args: list[str], name2value: Optional[Callable] = None, fallback: Optional[T] = None) -> 'DefaultDict':
        default, kwargs = DefaultDict.parse_configs(args, name2value=name2value, fallback=fallback)
        return DefaultDict(configs=kwargs, default=default)

    def get_first(self, keys, fallback=None):
        for key in keys:
            if key in self.configs:
                return self.configs[key]
        if self.default is not Unset:
            return self.default
        return fallback

    def __contains__(self, item):
        return item in self.configs or self.default is not Unset

    def __getitem__(self, item) -> T:
        if item not in self.configs and self.default is not Unset:
            return self.default
        else:
            return self.configs[item]

    def __bool__(self):
        return bool(self.configs) or self.default is not Unset



from inspect import getfullargspec
from typing import Collection


def decos(*decs):
    if len(decs) == 1 and isinstance(decs, Collection):
        decs = decs[0]

    def _fn(fn):
        for dec in reversed(decs):
            fn = dec(fn)
        return fn

    return _fn


def args(fn, kwargs):
    spec = getfullargspec(fn)
    args = spec.args
    return { k: v for k, v in kwargs.items() if k in args }



from utz.setup import setup

setup(
    name="utz",
    version="0.3.25",
    install_requires=open('requirements.txt', 'r').read(),
    extras_require={
        'pdf': [
            'reportlab',
            'pyPDF2',
        ],
        'setup': [
            'lxml',
            'mistune',
        ],
        'test': [
            'pytest',
            'pytest-mock',
        ],
    },
    url="https://github.com/runsascoded/utz",
)
